# Project Structure

This document describes the configuraitons file needed for Sable to process your
text files. Additional configuration files are described in subdirectories.

Most of these files will be YAML files. 

For more info on YAML, see the following links:
* https://yaml.org/spec/1.2/spec.html < Official YAML 1.2 spec.
* https://github.com/jbeder/yaml-cpp < library used to read YAML.

Sable uses the Asar library to assemble the binary + other ASM files
after conversion: https://github.com/RPGHacker/asar

## Main configuration file - config.yml

This is the primary config file, which defines the structure of the rest of the 
project. 

It MUST be named config.yml and MUST be in the same folder as the executable.

Config.yml has the following fields:

* files 
  * mainDir - The top level directory for all project files, except roms and 
configuration. The "main" asm file for each rom 
  * input
    * directory - The directory that contains the script files. 
  * output
    * directory - the subdirectory that will contain some asm. files generated by 
    Sable, as well as
    * binaries
      * mainDir - subdirectory of the main output directory that contains all
      binary files
      * textDir - subdirectory where Sable will place its converted binary text files.
    * fonts
      * dir - the subdirectory to write font width files to.
      * includes - additional files in the fonts subdirectory to include
    * extras - sequence of extra binary directories to include. Sable will add 
    `{output directory}/{binaries mainDir}/{tag of each extra}/{tag of eachextra}.asm` 
    to the main assembly file for each rom.
  * romDir - the directory that contains the main 
* config
  * directory - the subdirectory that contains the text conversion 
    configuration file.
  * inMapping - the name of the file(s) that holds the text configuration.
    * All files must be located in the config directory.
    * If there are more than one files that should be used, this field must be a sequence.
    Example:
    ```
    inMapping:
      - default.yml
      - menu.yml
      - somethingelse.yml
    ```
  * defaultMode - the default mode for text conversion. If no value is given, 
    the default is `normal`.
  * outputSize: the size of the output roms. Can be given as number of bytes, or 
    megabtyes/kilobytes with mb/kb suffixes. Max supported size is 8MB.
    * If 8MB is set as the output size, the actual output size will be more 
      like 7.9MB dues to technical limitations of SNES addressing.
  * mapper - the type of memory mapping the game uses.
    * Currenly supported values are `lorom`, `hirom`, `exlorom`, and `exhirom`.
      The default value is `lorom`.
    * If `lorom` or `hirom` are given, but outputSize is larger than 4m, sable will autocorrect 
      to use the expanded mapper format.
  * useMirrorBanks - whether to primarily use the mirrored (0x00 - 0x6F) banks for the LoROM mapper.
    * "true" or "false" are accepted values. The default is "false."
* roms - a sequence of all the input rom files to generate patches. Each should 
have the following fields:
  * name - the name of the output file, minus the extension(which is chosen 
  automatically based on the input file name).
  * header - optional. Can be set to true (assume file has a header), false 
  (assume file does not have a header), or auto (guess if the file is headered 
  based on the file size). Defaults to auto.
  * includes - optional. Additional files to be included specific to each rom. 
  Useful for defining revision specific addresses and the like.

### Example folder structure
```
sable.exe
asar.dll
config.yml
config
├── in.yml (inMapping config file)
project (mainDir)
├── text (input directory)
|   ├── folder1
|   |   ├── file01.txt
|   |   ├── file02.txt
|   |   ├── ...
|   |   ├── table.txt
|   ├── folder2
|   |   ├── ...
|   ├── ...
|   asm (output directory)
|   ├── someCustomCode.asm
|   ├── ...
|   ├── bin (binaries mainDir)
|   |   ├── text (binaries textDir)
|   |   |   ├── (bin files generated by sable)
|   |   ├── fonts
|   |   |   ├── fonts.asm (generated by sable)
|   |   |   ├── fontextras.asm
|   |   |   ├── ...
|   |   ├── custom 
|   |   |   ├── custom.asm
|   |   |   ├── custom1.bin
|   |   |   ├── custom2.bin
|   |   |   ├── ...
roms
├── rom1.smc
├── rom2.sfc
├── ...
```

## Input Mapping Configuration File

The configuration file that defines how text will be converted into binary.

This file is also a YAML file.

The file is a map of different possible text modes. The top-level tags are the
name of each text mode, which can be loaded either via the defaultMode setting
in the main configuration file, or in the individual script files. 

Note: the font name `default` is treated by Sable as an alias for the text in the
`defaultMode` setting, so it cannot be used as a tag name unless you also set 
the `defaultMode` setting to `default`.

Each font entry must have the following tags defined:
* Encoding: Defines how characters in the script will be converted to binary.
    * Each has the following fields:
        * code: Required. The binary value the character will be converted to.
        * length: The lengthin pixels of the character in the in-game font.
        This is used for line overflow calculation purposes, and may also be
        inserted into the game depending on other settings. If the font is
        fixed width, this can be omitted.
    * Helpful tip: You can use YAML anchors & aliases if you have multiple fonts
    that use the same encoding but have different properties.
    * If the Pages field is defined (
* Commands: Used for special non-character encodings - newlines, showing
portraits, closing windows, and the like. 
    * Each has the following fields:
        * code: Required. The binary value the command will be converted to.
        * newline: Optional. If defined(the value does not matter), Sable will 
        not automatically add a newline if a line break is reached after this
        command is read.
        * page: Optional. If set, sable will switch to the specified text page
        when this command is parsed. 
        Should be a number 0 or higher, but not higher than the number of pages + 1. 
    * Certain tags will be used by Sable for specific purposes:
        * End: The code defined for this command will automatically be defined 
        at the end of a text node, unless the "autoend" setting is disabled.
        * NewLine: Sable will insert the code for this command every time it
        reaches a line break in the text file it is parsing.
* HasDigraphs
    * Accepted values: `true` or `false`
    * Whether or not the current font has any multi-character encoding bytes.
* ByteWidth:
    * Accepted values: `1` or `2`
    * The number of bytes each character in the current font is encoded with.
    Then any instance of `text` will be parsed as `0x70,0x72,0x61,0x6E,0x6B` rather than whatever each individual character would be parse as otherwise.
        * in this case, the resulting string is 5 instead of 4 bytes, but this is more useful for cases where you want a smaller number of bytes.

Each text type entry may have the following additional tags:
* Extras
    * Simple encodings that can be used in brackets. Useful for arguments to commands.
* Nouns 
    * Multi-character strings that use a fixed encoding rather than parsing each character individually.
      * Most useful for faking variable width text in fixed width fonts.
      * For example, if you addd the following to a font:
      ```
      Nouns:
          text:
              code:
                  - 0x70
                  - 0x72
                  - 0x61
                  - 0x6E
                  - 0x6B
      ```
* Pages
    * A list of additional code pages which may be used during parsing.
      * This field must be a seuqence. There are two valid formats for each entry:
        * Each entry may be a map definig character encodings. In other words, 
        the same format as the "Encoding" field. An example:
        ```
        Pages:
          - A:
              code: 0x01
            B:
              code: 0x02
        ```
        * Each entry may also be a map with:
          * These required fields: 
            * An "Encoding" field, matching the same format used for the top-level Encoding field.
          * These optional fields:
            * A "Nouns" field matching the top-level Nouns field.
            * A "MaxEncodedValue" field, which has the same effect as the top-level 
            MaxEncodedValue field on a per-page basis.
          An example:
        ```
        Pages:
          - Encoding:
              A:
                code: 0x01
              B:
                code: 0x02
            MaxEncodedValue: 0x10
            Nouns:
              text:
                  code:
                      - 0x06
                      - 0x07
                      - 0x08
        ```
* CommandValue
    * If defined, this value will be inserted into the encoded text before 
    every command.
* MaxWidth:
    * The default maximum line width of the current font.
* FontWidthAddress:
    * Location to write font widths to. Can be either a hexaecimal SNES address,
    or an ASAR define.
* DefaultWidth
    * Must be a numeric value. If defined, it will be used as the width for all
    characters in the current font that do not have a specified.
* FixedWidth
    * Must be a scalar value. If defined, all characters in the current font will 
    have the same width.
        * If DefaultWidth is specified, that value is used.
        * Otherwise FixedWidth must be a numeric value, which 
        will be used as the width for all characters.
* MaxEncodedValue:
    * Used when writing font widths. If set, this will limit the number of font widths 
    which are written to the fotn file for the first page. 
    If this value is not defined, the maximum value will be calculated based on the 
    given byte width.
## Text file format

Text files should be grouped into subdirectories and placed inside the 
`textDir` subfolder in the project directory. 

Normal text files can be named whatever you like - as long as they're in the subfolders,
Sable will parse them. The exception is `table.txt` - if this file is present in a
subfolder, Sable will read this file for settings and parse the other files as entries
in a table.

### Normal text files

The following characters are reserved for special use:
* `#` - used for comments. All text after `#` on a line is ignored by Sable.
* `@` - used to define settings. 
    * @address - can be either `auto` or a hexadecimal address(`$` prefix optional).
    Using the auto option will assemble the text at whatever the last set address was,
    whether that's after the last text node or if it was set in a table file.
    If Sable attempts to parse text and no address has been set, it will stop parsing.
    * @type - the type of text, or the font if you prefer. Sable will look for this as a 
    top level tag in the inMapping yml file. If it's not found, Sable will stop parsing.
        * you can aslo use the option `default` to revert to the font set in `defaultMode`
    * @width - overrides the width setting used to calulate line overflows. 
        * Can be either a decimal value, or `off` to disable line overflow calculation.
    * @label - sets the label used for the next text node when generating binary and
    assembly files for Asar. 
        * The value given here should be used for the `entry` fields in table files.
        * If no label is defined, Sable will generate one based on the file name.
    * @autoend - Can be `on` or `off`, defaults to `on`.
        * If enabled, the `End` comamnd will be added to the converted text node once
        the end of the text file is reached. Additionally, it the `End` command is 
        read in the middle of a file, Sable will treat any future text as a new node.
    * @printpc - adds the `print pc` command to the text ASM file after the node is
    included. This will cause Asar to print the current address after the binary file.
    Useful for debugging purposes.
    * @page - causes the parser to switch to the specified code page. May be useful if
    a particular piece of text starts being read in something other than the default page.
      * The argument should be a decinal number and be within the number of defined pages.
      Examples:
        * `@page 0` - OK! Switches back to the default page.
        * `@page 1` - OK! Switches to the first page defined in the "Pages" setting.
* `[]` - Brackets can be used to enclose:
    * Commands
    * Extras
    * Hexadecimal data.
    
### Table files

If a subdirectory has a `table.txt` file, Sable will generate adiditonal binary information
based on the settings in the table file. This will be a table of addresses pointing to the
converted text.

Sable will automatically generate and Asar define for the table location based on the name
of the subdirectory: 
```
table_{directory name}
```

i.e. if the subdirectory is named `dialogue`, the location of the table will be defined as
`table_dialogue`.

Tablefiles require the following settings:
* address - SNES address for the location of the table.
* width - the length of the addresses as stored in the table in game. Can be either 2 or 3.
* file - Add a file in the subdirectory to parse for table data.
* entry - Defines the actual entries in the table. Uses the labels defined in normal text files.
    * Note: The number of entries and files in a table does not have to be the same - you can
    have multiple table entries in one text file.
    * If you have data inside the table that you aren't changing, you can use the `const`
    keyword instead of a label for a table entry. All the remaining text on the line will be
    copied directly to the ASM file.

The following settings may also be used:
* data - SNES address for the location of the data pointed to by the table.
    * Only needed if the encoded text needs to go somewhere other than where the table is located.
    If not defined, the data is written at the end of the table.
* savewidth - If this setting is given, Sable will encode the width of the text alongside the
address in the table.
